import pytest
import json
import os

import eth_keyfile

FIXTURES_FILE_PATH = os.path.join(
    os.path.dirname(os.path.dirname(eth_keyfile.__file__)),
    "fixtures",
    "KeyStoreTests",
    "basic_tests.json",
)

MEW_KEYFILE = {
    "json": {
        "Crypto": {
            "cipher": "aes-128-ctr",
            "cipherparams": {
                "iv": "7e7b02d2b4ef45d6c98cb885e75f48d5",
            },
            "ciphertext": "a7a5743a6c7eb3fa52396bd3fd94043b79075aac3ccbae8e62d3af94db00397c",  # noqa: E501
            "kdf": "scrypt",
            "kdfparams": {
                "dklen": 32,
                "n": 8192,
                "p": 1,
                "r": 8,
                "salt": "247797c7a357b707a3bdbfaa55f4c553756bca09fec20ddc938e7636d21e4a20",  # noqa: E501
            },
            "mac": "5a3ba5bebfda2c384586eda5fcda9c8397d37c9b0cc347fea86525cf2ea3a468",
        },
        "address": "0b6f2de3dee015a95d3330dcb7baf8e08aa0112d",
        "id": "3c8efdd6-d538-47ec-b241-36783d3418b9",
        "version": 3,
    },
    "password": "moomoocow",
    "priv": "21eac69b9a52f466bfe9047f0f21c9caf3a5cdaadf84e2750a9b3265d450d481",
}

# https://eips.ethereum.org/EIPS/eip-2335#test-cases
SCRYPT_KEYFILE_V4 = {
    "json": {
        "crypto": {
            "kdf": {
                "function": "scrypt",
                "params": {
                    "dklen": 32,
                    "n": 262144,
                    "p": 1,
                    "r": 8,
                    "salt": "d4e56740f876aef8c010b86a40d5f56745a118d0906a34e69aec8c0db1cb8fa3",  # noqa: E501
                },
                "message": "",
            },
            "checksum": {
                "function": "sha256",
                "params": {},
                "message": "d2217fe5f3e9a1e34581ef8a78f7c9928e436d36dacc5e846690a5581e8ea484",  # noqa: E501
            },
            "cipher": {
                "function": "aes-128-ctr",
                "params": {"iv": "264daa3f303d7259501c93d997d84fe6"},
                "message": "06ae90d55fe0a6e9c5c3bc5b170827b2e5cce3929ed3f116c2811e6366dfe20f",  # noqa: E501
            },
        },
        "description": "This is a test keystore that uses scrypt to secure the secret.",
        "pubkey": "9612d7a727c9d0a22e185a1c768478dfe919cada9266988cb32359c11f2b7b27f4ae4040902382ae2910c15e2b420d07",  # noqa: E501
        "path": "m/12381/60/3141592653/589793238",
        "uuid": "1d85ae20-35c5-4611-98e8-aa14a633906f",
        "version": 4,
    },
    "password": "testpassword🔑",
    "priv": "000000000019d6689c085ae165831e934ff763ae46a2a6c172b3f1b60a8ce26f",
}

PBKDF2_KEYFILE_V4 = {
    "json": {
        "crypto": {
            "kdf": {
                "function": "pbkdf2",
                "params": {
                    "dklen": 32,
                    "c": 262144,
                    "prf": "hmac-sha256",
                    "salt": "d4e56740f876aef8c010b86a40d5f56745a118d0906a34e69aec8c0db1cb8fa3",  # noqa: E501
                },
                "message": "",
            },
            "checksum": {
                "function": "sha256",
                "params": {},
                "message": "8a9f5d9912ed7e75ea794bc5a89bca5f193721d30868ade6f73043c6ea6febf1",  # noqa: E501
            },
            "cipher": {
                "function": "aes-128-ctr",
                "params": {"iv": "264daa3f303d7259501c93d997d84fe6"},
                "message": "cee03fde2af33149775b7223e7845e4fb2c8ae1792e5f99fe9ecf474cc8c16ad",  # noqa: E501
            },
        },
        "description": "This is a test keystore that uses PBKDF2 to secure the secret.",
        "pubkey": "9612d7a727c9d0a22e185a1c768478dfe919cada9266988cb32359c11f2b7b27f4ae4040902382ae2910c15e2b420d07",  # noqa: E501
        "path": "m/12381/60/0/0",
        "uuid": "64625def-3331-4eea-ab6f-782f3ed16a83",
        "version": 4,
    },
    "password": "testpassword🔑",
    "priv": "000000000019d6689c085ae165831e934ff763ae46a2a6c172b3f1b60a8ce26f",
}

with open(FIXTURES_FILE_PATH) as fixtures_file:
    BASE_KEYFILE_FIXTURES = json.load(fixtures_file)


KEYFILE_FIXTURES = BASE_KEYFILE_FIXTURES.copy()
KEYFILE_FIXTURES.update(
    {
        "MEW_generated_keyfile": MEW_KEYFILE,
        "scrypt_keyfile_v4": SCRYPT_KEYFILE_V4,
        "pbkdf2_keyfile_v4": PBKDF2_KEYFILE_V4,
    }
)


@pytest.fixture(params=KEYFILE_FIXTURES.keys())
def keyfile_data(request):
    return KEYFILE_FIXTURES[request.param]


@pytest.fixture
def v4_keyfile_data():
    return {"scrypt": SCRYPT_KEYFILE_V4, "pbkdf2": PBKDF2_KEYFILE_V4}



import pytest
import os

from eth_keyfile.exceptions import (
    EthKeyfileException,
)


def test_ethkeyfileexception_with_user_message():
    with pytest.raises(EthKeyfileException) as exception:
        raise EthKeyfileException(user_message="This failed!")
    assert exception.type is EthKeyfileException
    assert exception.value.user_message == "This failed!"


def test_ethkeyfileexception_with_kwargs():
    with pytest.raises(TypeError) as exception:
        raise EthKeyfileException(data={"message": "Unable to fulfill your request."})

    # For Python > 3.9, str exception includes 'EthKeyfileException.'
    expected = "__init__() got an unexpected keyword argument 'data'"
    actual = str(exception.value)
    assert exception.type is TypeError
    assert hasattr(exception.value, "data") is False
    assert expected in actual


def test_ethkeyfileexception_with_args():
    with pytest.raises(EthKeyfileException) as exception:
        raise EthKeyfileException("failed")
    assert exception.type is EthKeyfileException
    assert exception.value.user_message is None
    assert exception.value.args[0] == "failed"


ETH_KEYFILE_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "..", "eth_keyfile"
)
DEFAULT_EXCEPTIONS = (
    NotImplementedError,
    TypeError,
    ValueError,
)


def test_no_default_exceptions_are_raised_within_py_geth():
    for root, _dirs, files in os.walk(ETH_KEYFILE_PATH):
        for file in files:
            if file.endswith(".py"):
                file_path = os.path.join(root, file)
                with open(file_path, encoding="utf-8") as f:
                    for idx, line in enumerate(f):
                        for exception in DEFAULT_EXCEPTIONS:
                            exception_name = exception.__name__
                            if f"raise {exception_name}" in line:
                                raise Exception(
                                    f"``{exception_name}`` raised in eth-keyfile file "
                                    f"``{file}``, line {idx + 1}. "
                                    f"Replace with ``EthKeyfile{exception_name}``:\n"
                                    f"    file_path:{file_path}\n"
                                    f"    line:{idx + 1}"
                                )



from importlib.metadata import (
    version as __version,
)

from eth_keyfile.keyfile import (
    create_keyfile_json,
    decode_keyfile_json,
    extract_key_from_keyfile,
    load_keyfile,
)

__version__ = __version("eth-keyfile")



import hashlib
import hmac
import io
import json
from typing import (
    IO,
    Any,
    AnyStr,
    Literal,
    TypeVar,
    cast,
)
from unicodedata import (
    normalize,
)
from uuid import (
    uuid4,
)

from Crypto import (
    Random,
)
from Crypto.Cipher import (
    AES,
)
from Crypto.Hash import (
    SHA256,
)
from Crypto.Protocol.KDF import (
    scrypt,
)
from Crypto.Util import (
    Counter,
)
from eth_keys import (
    keys,
)
from eth_typing import (
    HexStr,
)
from eth_utils import (
    big_endian_to_int,
    decode_hex,
    encode_hex,
    is_dict,
    is_string,
    keccak,
    remove_0x_prefix,
    to_dict,
    to_int,
)
from py_ecc.bls12_381 import (
    curve_order,
)
from py_ecc.bls import (
    G2ProofOfPossession as bls,
)
from py_ecc.secp256k1 import (
    N,
)

from eth_keyfile.exceptions import (
    EthKeyfileNotImplementedError,
    EthKeyfileTypeError,
    EthKeyfileValueError,
)

UNICODE_CONTROL_CHARS = list(range(0x00, 0x20)) + list(range(0x7F, 0xA0))

# the maximum valid private key for secp256k1
MAX_V3_PRIVATE_KEY = N - 1

# the maximum valid private key for bls12-381
MAX_V4_PRIVATE_KEY = curve_order - 1

KDFType = Literal["pbkdf2", "scrypt"]
TKey = TypeVar("TKey")
TVal = TypeVar("TVal")


def encode_hex_no_prefix(value: AnyStr) -> HexStr:
    return remove_0x_prefix(encode_hex(value))


def load_keyfile(path_or_file_obj: str | IO[str]) -> Any:
    if is_string(path_or_file_obj):
        assert isinstance(path_or_file_obj, str)
        with open(path_or_file_obj) as keyfile_file:
            return json.load(keyfile_file)
    else:
        assert isinstance(path_or_file_obj, io.TextIOBase)
        return json.load(path_or_file_obj)


def create_keyfile_json(
    private_key: bytes | bytearray | memoryview,
    password: bytes,
    version: int = 3,
    kdf: KDFType = "pbkdf2",
    iterations: int | None = None,
    salt_size: int | None = None,
    description: str | None = None,
    path: str | None = None,
) -> dict[str, Any]:
    if version == 3:
        return _create_v3_keyfile_json(
            private_key, password, kdf, iterations, salt_size
        )
    if version == 4:
        return _create_v4_keyfile_json(
            private_key, password, kdf, iterations, salt_size, description, path
        )
    else:
        raise EthKeyfileNotImplementedError("Not yet implemented")


def decode_keyfile_json(raw_keyfile_json: dict[Any, Any], password: bytes) -> bytes:
    keyfile_json = normalize_keys(raw_keyfile_json)
    version = keyfile_json["version"]

    if version == 3:
        return _decode_keyfile_json_v3(keyfile_json, password)
    if version == 4:
        return _decode_keyfile_json_v4(keyfile_json, password)
    else:
        raise EthKeyfileNotImplementedError("Not yet implemented")


def extract_key_from_keyfile(path_or_file_obj: str | IO[str], password: bytes) -> bytes:
    keyfile_json = load_keyfile(path_or_file_obj)
    private_key = decode_keyfile_json(keyfile_json, password)
    return private_key


@to_dict
def normalize_keys(keyfile_json: dict[Any, Any]) -> Any:
    for key, value in keyfile_json.items():
        if is_string(key):
            norm_key = key.lower()
        else:
            norm_key = key

        if is_dict(value):
            norm_value = normalize_keys(value)
        else:
            norm_value = value

        yield norm_key, norm_value


#
# Version 3 creator
#
DKLEN = 32
SCRYPT_R = 8
SCRYPT_P = 1


def _create_v3_keyfile_json(
    private_key: bytes | bytearray | memoryview,
    password: bytes,
    kdf: KDFType,
    work_factor: int | None = None,
    salt_size: int | None = None,
) -> dict[str, Any]:
    if work_factor is None:
        work_factor = get_default_work_factor_for_kdf(kdf)
    if salt_size is None:
        salt_size = 16

    if to_int(private_key) > MAX_V3_PRIVATE_KEY:
        raise EthKeyfileValueError(
            "Invalid `private_key`, exceeds maximum valid secp256k1 key "
            f"value of {MAX_V3_PRIVATE_KEY}"
        )

    salt = Random.get_random_bytes(salt_size)

    if kdf == "pbkdf2":
        derived_key = _pbkdf2_hash(
            password,
            hash_name="sha256",
            salt=salt,
            iterations=work_factor,
            dklen=DKLEN,
        )
        kdfparams = {
            "c": work_factor,
            "dklen": DKLEN,
            "prf": "hmac-sha256",
            "salt": encode_hex_no_prefix(salt),
        }
    elif kdf == "scrypt":
        derived_key = _scrypt_hash(
            password,
            salt=salt,
            buflen=DKLEN,
            r=SCRYPT_R,
            p=SCRYPT_P,
            n=work_factor,
        )
        kdfparams = {
            "dklen": DKLEN,
            "n": work_factor,
            "r": SCRYPT_R,
            "p": SCRYPT_P,
            "salt": encode_hex_no_prefix(salt),
        }
    else:
        raise EthKeyfileNotImplementedError(f"KDF not implemented: {kdf}")

    iv_bytes = Random.get_random_bytes(16)
    encrypt_key = derived_key[:16]
    ciphertext = encrypt_aes_ctr(private_key, encrypt_key, big_endian_to_int(iv_bytes))
    mac = keccak(derived_key[16:32] + ciphertext)

    address = keys.PrivateKey(private_key).public_key.to_checksum_address()

    return {
        "address": remove_0x_prefix(address),
        "crypto": {
            "cipher": "aes-128-ctr",
            "cipherparams": {
                "iv": iv_bytes.hex(),
            },
            "ciphertext": encode_hex_no_prefix(ciphertext),
            "kdf": kdf,
            "kdfparams": kdfparams,
            "mac": encode_hex_no_prefix(mac),
        },
        "id": str(uuid4()),
        "version": 3,
    }


#
# Verson 3 decoder
#
def _decode_keyfile_json_v3(keyfile_json: dict[str, Any], password: bytes) -> bytes:
    crypto = keyfile_json["crypto"]
    kdf = crypto["kdf"]

    # Derive the encryption key from the password using the key derivation
    # function.
    if kdf == "pbkdf2":
        derived_key = _derive_pbkdf_key(crypto["kdfparams"], password)
    elif kdf == "scrypt":
        derived_key = _derive_scrypt_key(crypto["kdfparams"], password)
    else:
        raise EthKeyfileTypeError(f"Unsupported key derivation function: {kdf}")

    # Validate that the derived key matchs the provided MAC
    ciphertext = decode_hex(crypto["ciphertext"])
    mac = keccak(derived_key[16:32] + ciphertext)

    expected_mac = decode_hex(crypto["mac"])

    if not hmac.compare_digest(mac, expected_mac):
        raise EthKeyfileValueError("MAC mismatch")

    # Decrypt the ciphertext using the derived encryption key to get the
    # private key.
    encrypt_key = derived_key[:16]
    cipherparams = crypto["cipherparams"]
    iv = big_endian_to_int(decode_hex(cipherparams["iv"]))

    private_key = decrypt_aes_ctr(ciphertext, encrypt_key, iv)

    return private_key


#
# Version 4 creator
#


def _create_v4_keyfile_json(
    private_key: bytes | bytearray | memoryview,
    password: bytes,
    kdf: KDFType,
    work_factor: int | None = None,
    salt_size: int | None = None,
    description: str | None = None,
    path: str | None = None,
) -> dict[str, Any]:
    if work_factor is None:
        work_factor = get_default_work_factor_for_kdf(kdf)
    if salt_size is None:
        salt_size = 32
    if description is None:
        description = ""
    if path is None:
        path = ""

    aes_iv = Random.get_random_bytes(16)

    if to_int(private_key) > MAX_V4_PRIVATE_KEY:
        raise EthKeyfileValueError(
            "Invalid `private_key`, exceeds maximum valid bls12-381 key "
            f"value of {MAX_V4_PRIVATE_KEY}"
        )

    salt: bytes = Random.get_random_bytes(salt_size)
    uuid: str = str(uuid4())

    # clean password
    password_str = normalize("NFKD", password.decode())
    password_str = "".join(
        c for c in password_str if ord(c) not in UNICODE_CONTROL_CHARS
    )
    clean_password = password_str.encode("UTF-8")

    if kdf == "pbkdf2":
        derived_key = _pbkdf2_hash(
            password=clean_password,
            hash_name="sha256",
            salt=salt,
            iterations=work_factor,
            dklen=DKLEN,
        )
        kdfparams = {
            "c": work_factor,
            "dklen": DKLEN,
            "prf": "hmac-sha256",
            "salt": encode_hex_no_prefix(salt),
        }
    elif kdf == "scrypt":
        derived_key = _scrypt_hash(
            password=clean_password,
            salt=salt,
            buflen=DKLEN,
            r=SCRYPT_R,
            p=SCRYPT_P,
            n=work_factor,
        )
        kdfparams = {
            "dklen": DKLEN,
            "n": work_factor,
            "r": SCRYPT_R,
            "p": SCRYPT_P,
            "salt": encode_hex_no_prefix(salt),
        }
    else:
        raise EthKeyfileNotImplementedError(f"KDF not implemented: {kdf}")

    encrypt_key = derived_key[:16]
    encoded_pk = encrypt_aes_ctr(private_key, encrypt_key, big_endian_to_int(aes_iv))
    checksum_msg = SHA256.new(derived_key[16:32] + encoded_pk)

    kdf_key = {
        "function": kdf,
        "params": kdfparams,
        "message": "",
    }

    cipher_key = {
        "function": "aes-128-ctr",
        "params": {
            "iv": aes_iv.hex(),
        },
        "message": encoded_pk.hex(),
    }

    checksum_key = {
        "function": "sha256",
        "params": {},
        "message": checksum_msg.hexdigest(),
    }

    return {
        "crypto": {
            "kdf": kdf_key,
            "checksum": checksum_key,
            "cipher": cipher_key,
        },
        "description": description,
        "pubkey": bls.SkToPk(big_endian_to_int(private_key)).hex(),
        "path": path,
        "uuid": uuid,
        "version": 4,
    }


#
# Version 4 decoder
#
def _decode_keyfile_json_v4(keyfile_json: dict[str, Any], password: bytes) -> bytes:
    crypto = keyfile_json["crypto"]
    kdf = crypto["kdf"]["function"]

    # Derive the encryption key from the password using the key derivation
    # function.
    if kdf == "pbkdf2":
        derived_key = _derive_pbkdf_key(crypto["kdf"]["params"], password)
    elif kdf == "scrypt":
        derived_key = _derive_scrypt_key(crypto["kdf"]["params"], password)
    else:
        raise EthKeyfileTypeError(f"Unsupported key derivation function: {kdf}")

    cipher_message = decode_hex(crypto["cipher"]["message"])
    checksum_message = crypto["checksum"]["message"]

    if (
        hashlib.sha256(derived_key[16:32] + cipher_message).hexdigest()
        != checksum_message
    ):
        raise EthKeyfileValueError("Checksum mismatch")

    # Decrypt the cipher message using the derived encryption key to get the
    # private key.
    encrypt_key = derived_key[:16]
    cipherparams = crypto["cipher"]["params"]
    iv = big_endian_to_int(decode_hex(cipherparams["iv"]))

    private_key = decrypt_aes_ctr(cipher_message, encrypt_key, iv)

    return private_key


#
# Key derivation
#
def _derive_pbkdf_key(kdf_params: dict[str, Any], password: bytes) -> bytes:
    salt = decode_hex(kdf_params["salt"])
    dklen = kdf_params["dklen"]
    should_be_hmac, _, hash_name = kdf_params["prf"].partition("-")
    assert should_be_hmac == "hmac"
    iterations = kdf_params["c"]

    derive_pbkdf_key = _pbkdf2_hash(password, hash_name, salt, iterations, dklen)

    return derive_pbkdf_key


def _derive_scrypt_key(kdf_params: dict[str, Any], password: bytes) -> bytes:
    salt = decode_hex(kdf_params["salt"])
    p = kdf_params["p"]
    r = kdf_params["r"]
    n = kdf_params["n"]
    buflen = kdf_params["dklen"]

    derived_scrypt_key = _scrypt_hash(
        password,
        salt=salt,
        n=n,
        r=r,
        p=p,
        buflen=buflen,
    )
    return derived_scrypt_key


def _scrypt_hash(
    password: bytes, salt: bytes, n: int, r: int, p: int, buflen: int
) -> bytes:
    derived_key = scrypt(
        # scrypt uniquely accepts str, not bytes, for password and salt
        # latin-1 is used in pycryptodome for bytes encoding/decoding
        password.decode("latin-1"),
        salt.decode("latin-1"),
        key_len=buflen,
        N=n,
        r=r,
        p=p,
        num_keys=1,
    )
    # num_keys is set to 1, so rtype will always be bytes, not Tuple[bytes, ...]
    return cast(bytes, derived_key)


def _pbkdf2_hash(
    password: bytes, hash_name: str, salt: bytes, iterations: int, dklen: int
) -> bytes:
    derived_key = hashlib.pbkdf2_hmac(
        hash_name=hash_name,
        password=password,
        salt=salt,
        iterations=iterations,
        dklen=dklen,
    )

    return derived_key


#
# Encryption and Decryption
#
def decrypt_aes_ctr(ciphertext: bytes, key: bytes, iv: int) -> bytes:
    ctr = Counter.new(128, initial_value=iv, allow_wraparound=True)
    encryptor = AES.new(key, AES.MODE_CTR, counter=ctr)
    return encryptor.decrypt(ciphertext)


def encrypt_aes_ctr(
    value: bytes | bytearray | memoryview, key: bytes, iv: int
) -> bytes:
    ctr = Counter.new(128, initial_value=iv, allow_wraparound=True)
    encryptor = AES.new(key, AES.MODE_CTR, counter=ctr)
    ciphertext = encryptor.encrypt(value)
    return ciphertext


#
# Utility
#
def get_default_work_factor_for_kdf(kdf: KDFType) -> int:
    if kdf == "pbkdf2":
        return 1000000
    elif kdf == "scrypt":
        return 262144
    else:
        raise EthKeyfileValueError(f"Unsupported key derivation function: {kdf}")



from typing import (
    Any,
)


class EthKeyfileException(Exception):
    """
    Exception mixin inherited by all exceptions of eth-keyfile

    This allows::

        try:
            some_call()
        except EthKeyfileException:
            # deal with eth-keyfile exception
        except:
            # deal with other exceptions
    """

    user_message: str | None = None

    def __init__(
        self,
        *args: Any,
        user_message: str | None = None,
    ):
        super().__init__(*args)

        # Assign properties of EthKeyfileException
        self.user_message = user_message


class EthKeyfileNotImplementedError(EthKeyfileException, NotImplementedError):
    """
    An eth-keyfile exception wrapper for `NotImplementedError`, for better control over
    exception handling
    """


class EthKeyfileValueError(EthKeyfileException, ValueError):
    """
    An eth-keyfile exception wrapper for `ValueError`, for better control over
    exception handling
    """


class EthKeyfileTypeError(EthKeyfileException, TypeError):
    """
    An eth-keyfile exception wrapper for `TypeError`, for better control over
    exception handling
    """



import pytest
import json

from eth_utils import (
    decode_hex,
)

from eth_keyfile.keyfile import (
    create_keyfile_json,
    extract_key_from_keyfile,
    load_keyfile,
)

PRIVATE_KEY = decode_hex(
    "7a28b5ba57c53603b0b07b56bba752f7784bf506fa95edc395f5cf6c7514fe9d"
)
PASSWORD = b"foo"


@pytest.fixture(params=["pbkdf2", "scrypt"])
def keyfile_json(request):
    _keyfile_json = create_keyfile_json(
        PRIVATE_KEY, PASSWORD, kdf=request.param, iterations=2
    )
    return _keyfile_json


@pytest.fixture()
def keyfile_path(tmpdir, keyfile_json):
    _keyfile_path = tmpdir.join("keyfile.json")
    _keyfile_path.write(json.dumps(keyfile_json))
    return str(_keyfile_path)


def test_load_keyfile_with_file_obj(keyfile_path, keyfile_json):
    with open(keyfile_path) as keyfile_file:
        loaded_keyfile_json = load_keyfile(keyfile_file)

    assert loaded_keyfile_json == keyfile_json


def test_load_keyfile_with_path(keyfile_path, keyfile_json):
    loaded_keyfile_json = load_keyfile(keyfile_path)

    assert loaded_keyfile_json == keyfile_json


def test_extract_key_from_keyfile_with_file_obj(keyfile_path):
    with open(keyfile_path) as keyfile_file:
        private_key = extract_key_from_keyfile(keyfile_file, PASSWORD)

    assert private_key == PRIVATE_KEY


def test_extract_key_from_keyfile_with_path(keyfile_path):
    private_key = extract_key_from_keyfile(keyfile_path, PASSWORD)

    assert private_key == PRIVATE_KEY



import pytest

from eth_utils import (
    decode_hex,
    to_bytes,
)

from eth_keyfile.exceptions import (
    EthKeyfileValueError,
)
from eth_keyfile.keyfile import (
    MAX_V3_PRIVATE_KEY,
    MAX_V4_PRIVATE_KEY,
    create_keyfile_json,
    decode_keyfile_json,
)

PRIVATE_KEY = decode_hex(
    "7a28b5ba57c53603b0b07b56bba752f7784bf506fa95edc395f5cf6c7514fe9d"
)
PASSWORD = b"foo"

PRIVATE_KEYS_VALID_FOR_3_AND_4 = [
    decode_hex("000000000019d6689c085ae165831e934ff763ae46a2a6c172b3f1b60a8ce26f"),
    decode_hex("21eac69b9a52f466bfe9047f0f21c9caf3a5cdaadf84e2750a9b3265d450d481"),
    to_bytes(MAX_V4_PRIVATE_KEY),
]
PRIVATE_KEYS_VALID_FOR_3_NOT_4 = [
    decode_hex("7a28b5ba57c53603b0b07b56bba752f7784bf506fa95edc395f5cf6c7514fe9d"),
    decode_hex("deadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeef"),
]

PRIVATE_KEYS_INVALID_FOR_3_AND_4 = [
    decode_hex("ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff"),
    to_bytes(MAX_V3_PRIVATE_KEY + 1),
]

VALID_V3 = PRIVATE_KEYS_VALID_FOR_3_AND_4 + PRIVATE_KEYS_VALID_FOR_3_NOT_4
VALID_V4 = PRIVATE_KEYS_VALID_FOR_3_AND_4
INVALID_V3 = PRIVATE_KEYS_INVALID_FOR_3_AND_4
INVALID_V4 = PRIVATE_KEYS_VALID_FOR_3_NOT_4 + PRIVATE_KEYS_INVALID_FOR_3_AND_4


@pytest.mark.parametrize("private_key", VALID_V3)
@pytest.mark.parametrize("kdf", ["pbkdf2", "scrypt"])
@pytest.mark.parametrize("iterations", [1, 2, 4])
def test_keyfile_v3_creation(private_key, kdf, iterations):
    keyfile_json = create_keyfile_json(
        private_key,
        password=PASSWORD,
        version=3,
        kdf=kdf,
        iterations=iterations,
    )
    derived_private_key = decode_keyfile_json(keyfile_json, PASSWORD)
    assert derived_private_key == private_key


@pytest.mark.parametrize("private_key", VALID_V4)
@pytest.mark.parametrize("kdf", ["pbkdf2", "scrypt"])
@pytest.mark.parametrize("iterations", [1, 2, 4])
@pytest.mark.parametrize("salt_size", [1, 8, 9, 64, 128])
@pytest.mark.parametrize("description", ["", "foo", "bar"])
@pytest.mark.parametrize("path", ["a/b", "m/12381/60/3141592653/589793238"])
def test_keyfile_v4_creation(
    private_key, kdf, iterations, salt_size, description, path
):
    keyfile_json = create_keyfile_json(
        private_key,
        password=PASSWORD,
        version=4,
        kdf=kdf,
        iterations=iterations,
        salt_size=salt_size,
        description=description,
        path=path,
    )
    derived_private_key = decode_keyfile_json(keyfile_json, PASSWORD)
    assert derived_private_key == private_key


@pytest.mark.parametrize("private_key", INVALID_V3)
def test_invalid_v3_private_key_raises(private_key):
    with pytest.raises(EthKeyfileValueError):
        create_keyfile_json(
            private_key,
            password=PASSWORD,
            version=3,
        )


@pytest.mark.parametrize("private_key", INVALID_V4)
def test_invalid_v4_private_key_raises(private_key):
    with pytest.raises(EthKeyfileValueError):
        create_keyfile_json(
            private_key,
            password=PASSWORD,
            version=4,
        )


def test_pbkdf2_keyfile_salt32_creation():
    keyfile_json = create_keyfile_json(
        PRIVATE_KEY,
        password=PASSWORD,
        kdf="pbkdf2",
        iterations=1,
        salt_size=32,
    )
    assert len(keyfile_json["crypto"]["kdfparams"]["salt"]) == 32 * 2
    derived_private_key = decode_keyfile_json(keyfile_json, PASSWORD)
    assert derived_private_key == PRIVATE_KEY


def test_scrypt_keyfile_address():
    keyfile_json = create_keyfile_json(
        PRIVATE_KEY,
        password=PASSWORD,
        kdf="scrypt",
        iterations=2,
    )
    assert keyfile_json["address"] == "008AeEda4D805471dF9b2A5B0f38A0C3bCBA786b"


def recursive_check_dicts(dict1, dict2):
    for key in dict1:
        if isinstance(dict1[key], dict):
            recursive_check_dicts(dict1[key], dict2[key])
        else:
            assert type(dict1[key]) is type(dict2[key])


@pytest.mark.parametrize("kdf", ["pbkdf2", "scrypt"])
def test_v4_create_datatypes_match(kdf, v4_keyfile_data):
    test_pk = v4_keyfile_data[kdf]["priv"]
    test_pw = v4_keyfile_data[kdf]["password"]
    new_keyfile = create_keyfile_json(
        private_key=decode_hex(test_pk),
        password=test_pw.encode("utf-8"),
        version=4,
        kdf=kdf,
        description=v4_keyfile_data[kdf]["json"]["description"],
        path=v4_keyfile_data[kdf]["json"]["path"],
    )
    recursive_check_dicts(new_keyfile, v4_keyfile_data[kdf]["json"])



import codecs

from eth_utils import (
    decode_hex,
)

from eth_keyfile.keyfile import (
    decode_keyfile_json,
)


def test_decoding_keyfile(keyfile_data):
    password = codecs.encode(keyfile_data["password"], "utf8")
    keyfile_json = keyfile_data["json"]
    private_key = keyfile_data["priv"]

    derived_private_key = decode_keyfile_json(keyfile_json, password)
    assert decode_hex(private_key) == derived_private_key



import doctest
import re


def extract_doctests(md_content):
    """Extract Python code blocks that appear to be doctests."""
    # This regex matches code blocks that start with '```python' and end with '```'
    code_blocks = re.findall(r"```python(.*?)```", md_content, re.DOTALL)
    # Concatenate all code blocks into a single string of code
    return "\n\n".join(code_blocks)


def run_doctests(code):
    """Run doctests in the given block of code."""
    parser = doctest.DocTestParser()
    test = parser.get_doctest(code, {}, "extracted_doctests", "extracted", 0)
    runner = doctest.DocTestRunner()
    runner.run(test)


with open("README.md") as f:
    markdown_content = f.read()

code_to_test = extract_doctests(markdown_content)
run_doctests(code_to_test)


def test_readme_examples():
    run_doctests(code_to_test)
